import datetime
import json
import os
import re
import shutil
import smtplib
import subprocess
import update_package.common as common
from email.mime.text import MIMEText

class RequestException(Exception):
    pass


class RequestLog(object):
    """
    Manages logging progress of requests
    """
    def __init__(self, request):
        without_ext, ext = os.path.splitext(request)
        self.request_id = os.path.basename(without_ext)
        self.log_fname = os.path.join('results', self.request_id + '.log')
        self.log = open(self.log_fname, "w", 0)
        self.log_message("Started processing " + self.request_id)

    def log_message(self, msg):
        """
        Record process to log file and stdout
        """
        timestamp = datetime.datetime.utcnow().isoformat()
        self.log.write("%s: %s\n" % (timestamp, msg))
        print msg

    def close(self):
        """
        Finalize the request log
        """
        self.log_message("Request complete")
        self.log.close()
        subprocess.check_call(['git', 'add', self.log_fname])


def setup_repo(repo):
    """
    Initialize the repository containing requests
    """
    without_ext, _ = os.path.splitext(repo)
    repo_dir = os.path.basename(without_ext)
    shutil.rmtree(repo_dir, ignore_errors=True)

    subprocess.check_call(['git', 'clone', repo])
    results_dir = os.path.join(repo_dir, 'results')
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)

    return repo_dir


def email_alert(args, subject, content):
    if not args.mail_alert_to:
        return
    passphrase = os.environ.get('UPDATE_MAIL_PASSWORD')

    msg = MIMEText(content)
    msg['Subject'] = subject
    msg['To'] = args.mail_alert_to
    if args.mail_from:
        msg['From'] = args.mail_from

    s = smtplib.SMTP(args.smtp_server)
    if args.smtp_user and passphrase:
        s.login(args.smtp_user, passphrase)
    s.sendmail(None, msg['To'], msg.as_string())
    s.quit()


def report_progress(logger, args, subject, content):
    """
    Report progress to log file and via email
    """
    logger.log_message("%s:\n%s" % (subject, content))
    email_alert(args, subject + " [%s]" % logger.request_id, content)


def decrypt_request(request_signed):
    """
    Verify signature of request and decrypt
    """
    without_ext, ext = os.path.splitext(request_signed)
    request_base = os.path.basename(without_ext)
    request = os.path.join('results', request_base)

    subprocess.check_call(['git', 'mv', request_signed, 'results'])
    ret, out = common.call_stderr(['gpg', '--output', request, '--decrypt',
                            os.path.join('results', request_base + ext)])
    if ret != 0:
        raise RequestException("Failed to decrypt request\n" + out)

    match = re.search(r'Good signature from "(.*)"', out)
    committer = match.group(1) if match else "*** UNKNOWN ***"

    return committer, request


def check_request(args, request, required_keys=[]):
    """
    Check request is valid, return parsed data
    """
    try:
        with open(request) as fileh:
            jdata = json.load(fileh)
    except ValueError as err:
        raise RequestException(str(err))

    if not isinstance(jdata, dict):
        raise RequestException("Request is not a JSON object")

    for key in required_keys:
        if key not in jdata:
            raise RequestException("Missing key: " + key)

    return jdata


def copy_file(copy_from, copy_to):
    """
    Copy a single file
    """
    try:
        common.in_place_copy(copy_from, copy_to)
    except subprocess.CalledProcessError:
        raise RequestException("Failed to perform in-place copy of %s to %s" %
                               (copy_from, copy_to))
    except ValueError as err:
        raise RequestException(str(err))


def copy_dir(copy_from, copy_to, dest):
    """
    Copy a directory
    """
    try:
        common.in_place_copy(copy_from, copy_to, dest)
    except subprocess.CalledProcessError:
        raise RequestException("Failed to perform in-place copy of %s to %s" %
                               (copy_from, os.path.join(copy_to, dest)))
    except ValueError as err:
        raise RequestException(str(err))


