import argparse
import glob
import os
import os.path
import shutil
import subprocess
import sys
import tempfile
import time

from update_package.common import *
from update_package.server_common import *


def fetch_pub_key(args, key, key_path):
    """
    Fetch the public key for a given key name
    """
    cmd = 'gpg'
    if args.gpg_binary:
        cmd = args.gpg_binary
    ret, err = call_stderr([cmd, '--output', key_path, '--armor',
                            '--export', key])
    if ret != 0:
        raise RequestException("Failed to fetch key " + key + "\n" + err)


def fetch_pub_keys(args, keyfiles, output_dir):
    """
    Fetch public keys and build a mapping from key name
    """
    keymap = {}
    for keyfile in keyfiles:
        print "Fetch %(key)s to %(filename)s" % keyfile
        key_path = os.path.join(output_dir, keyfile['filename'])
        fetch_pub_key(args, keyfile['key'], key_path)
        keymap[keyfile['key']] = key_path
    return keymap


def sign_media(media_url, key, keymap, output_dir):
    """
    Sign installation media
    """
    base_name = os.path.basename(media_url)
    pubkeys = []
    for keyfile in keymap.values():
        pubkeys += ['--keyfile', keyfile]

    ret, err = call_stderr(['sign-media', '--no-passphrase', '--input',
                            media_url, '--key', key, '--output',
                            os.path.join(output_dir, base_name)] + pubkeys)
    if ret != 0:
        raise RequestException("Failed to sign media %s\n%s" %
                               (media_url, err))


def sign_update(update, key, keymap, output_dir):
    """
    Sign an update
    """
    ret, err = call_stderr(['sign-update', '--no-passphrase', '--output',
                            output_dir, '--key', key, '--keyfile', keymap[key],
                            update])
    if ret != 0:
        raise RequestException("Failed to sign update %s\n%s" % (update, err))


def sign_source(source, key, keymap, output_dir):
    """
    Sign source image
    """
    base_name = os.path.basename(source)
    ret, err = call_stderr(['sign-source', '--no-passphrase', '--input',
                            source, '--output',
                            os.path.join(output_dir, base_name), '--key', key,
                            '--keyfile',  keymap[key]])
    if ret != 0:
        raise RequestException("Failed to sign source %s\n%s" % (source, err))


def upload_release(output_dir, media_url, logger):
    """
    Upload signed artefacts
    """
    try:
        for local_path in glob.glob(output_dir + '/*.iso'):
            logger.log_message("Uploading %s to %s" % (local_path, media_url))
            deploy_file(local_path, media_url)
    except subprocess.CalledProcessError:
        raise RequestException("Failed to upload release to " + media_url)


def process_request(args, jdata, logger):
    """
    Act on valid request
    """
    output_dir = tempfile.mkdtemp(prefix='sign-media-')
    release_dir = os.path.join(output_dir, 'release')
    os.mkdir(release_dir)
    try:
        keymap = fetch_pub_keys(args, jdata['keyfiles'], output_dir)
        if 'media' in jdata:
            logger.log_message("Starting signing " + jdata['media'])
            sign_media(jdata['media'], jdata['key'], keymap, release_dir)
            report_progress(logger, args, "Sign media",
                            "Signing of %s completed" % jdata['media'])
        if 'updates' in jdata:
            for update in jdata['updates']:
                logger.log_message("Starting signing " + update)
                sign_update(update, jdata['key'], keymap, release_dir)
                report_progress(logger, args, "Sign update",
                                "Signing of %s completed" % update)
        if 'sources' in jdata:
            for source in jdata['sources']:
                logger.log_message("Starting signing " + source)
                sign_source(source, jdata['key'], keymap, release_dir)
                report_progress(logger, args, "Sign source",
                                "Signing of %s completed" % source)

        upload_release(release_dir, jdata['release'], logger)

        if 'copy_files' in jdata:
            for cfile in jdata['copy_files']:
                logger.log_message("Copying single file " + cfile)
                copy_file(cfile, jdata['release'])

        if 'copy_dirs' in jdata:
            for cdir in jdata['copy_dirs']:
                logger.log_message("Copying directory %s to %s" %
                                   (cdir['url'], cdir['release_path']))
                copy_dir(cdir['url'], jdata['release'], cdir['release_path'])

    finally:
        shutil.rmtree(output_dir, ignore_errors=True)

    return "Uploaded release to " + jdata['release']


def poll_repo(args):
    """
    Check for new signing requests and act on them
    """
    ret = subprocess.call(['git', 'pull', '-q'])
    if ret != 0:
        print "Unable to pull remote repository"
        return

    requests = glob.glob('requests/*.json.signed')

    for request_signed in requests:
        try:
            logger = RequestLog(request_signed)
            try:
                committer, request = decrypt_request(request_signed)

                # validly signed request received
                logger.log_message("Valid signing request from " + committer)
                email_alert(args, "Accepted: " + logger.request_id,
                            "Started processing signing request from " +
                            committer)
                required_keys = ['key', 'keyfiles', 'release']
                request_json = check_request(args, request, required_keys=required_keys)
                content = process_request(args, request_json, logger)
                subject = "Signing request completed"
            except RequestException as err:
                content = str(err)
                subject = "Signing request failed"

            report_progress(logger, args, subject, content)
        finally:
            logger.close()
        subprocess.check_call(['git', 'commit', '-a', '-m', 'Processed ' +
                               request_signed])

    if len(requests) > 0:
        retries = 5

        while retries != 0:
            ret = subprocess.call(['git', 'pull', '-r'])
            if ret == 0:
                ret = subprocess.call(['git', 'push'])
                if ret == 0:
                    break
            print "Failed to push results, retrying"
            retries -= 1
            time.sleep(1)

        if retries == 0:
            print "Unable to push results to remote repository"


def parse_args_or_exit(argv=None):
    """
    Parse command line options
    """
    parser = argparse.ArgumentParser(description="Sign product builds")
    parser.add_argument("--repo", required=True,
                        help="git repository to poll")
    parser.add_argument("--gpg-server", default=None,
                        help="Remote GnuPG signing service")
    parser.add_argument("--gpg-user", default=None,
                        help="Remote shell account")
    parser.add_argument("--gpg-binary", default=None,
                        help="Executable to use for signing")
    parser.add_argument("--poll-interval", type=int, default=60,
                        help="Polling interval in seconds")
    parser.add_argument("--smtp-server", default=None,
                        help="smtp server for email alert")
    parser.add_argument("--smtp-user", default=None,
                        help="smtp user to logon")
    parser.add_argument("--mail-alert-to", default=None,
                        help="Mail alert group send to")
    parser.add_argument("--mail-from", metavar="ADDR",
                        help="Mail sender")

    return parser.parse_args(argv)


def cleanup():
    """
    Delete any temporary files from a previous run
    """
    try:
        tmpdir = tempfile.gettempdir()
    except IOError:
        # no temp dir, filesystem probably full
        tmpdir = '/tmp'

    dirs = [os.path.join(tmpdir, d) for d in ('fetch-file-*', 'sign-*')]
    subprocess.call("rm -rf " + ' '.join(dirs), shell=True)


def main(argv):
    """
    Poll for signing requests and act on them
    """
    args = parse_args_or_exit(argv)
    if args.gpg_server:
        os.environ['UPDATE_GPG_SERVER'] = args.gpg_server
    if args.gpg_user:
        os.environ['UPDATE_GPG_USER'] = args.gpg_user
    if args.gpg_binary:
        os.environ['UPDATE_GPG_BIN'] = args.gpg_binary

    cleanup()
    repo_dir = setup_repo(args.repo)
    os.chdir(repo_dir)

    while True:
        poll_repo(args)
        time.sleep(args.poll_interval)


def _main():
    """
    Entry point for setuptools CLI wrapper
    """
    main(sys.argv[1:])


# Entry point when run directly
if __name__ == "__main__":
    _main()
